/**
 * 
 */
package de.pdark.decentxml;

import de.pdark.decentxml.XMLTokenizer.Type;

public class Attribute extends BasicNode
{
    private String name;
    private String value;
    private char quoteChar;
    
    public Attribute (XMLTokenizer.Token token)
    {
        super (token);
        
        String s = token.getText ();
        int pos = s.indexOf ('=');
        
        this.name = s.substring (0, pos);
        this.quoteChar = s.charAt (pos+1);
        this.value = s.substring (pos+2, s.length () - 1);
    }
    
    public Attribute (String name, String value)
    {
        super (Type.ATTRIBUTE, name + "=" + quotedAttributeValue (value));
        
        this.name = name;
        this.value = value;
    }

    public Attribute (String name, String value, char quoteChar)
    {
        super (Type.ATTRIBUTE, name + "=" + quoteChar + XMLParser.escapeXMLText (value) + quoteChar);
        this.quoteChar = quoteChar;
    }
    
    private static String quotedAttributeValue (String value)
    {
        if (value.indexOf ('"') == -1)
            return "\"" + XMLParser.escapeXMLText (value) + "\"";
        if (value.indexOf ('\'') == -1)
            return "'" + XMLParser.escapeXMLText (value) + "'";
        
        return "\"" + XMLParser.escapeXMLText (value).replaceAll ("\"", "&quot;") + "\"";
    }
    
    public String getName ()
    {
        return name;
    }
    
    public void setName (String name)
    {
        this.name = name;
        
        updateValue ();
    }
    
    private void updateValue ()
    {
        if (quoteChar == 0)
            setValue (name + "=" + quotedAttributeValue (value));
        else
            setValue (name + "=" + quoteChar + XMLParser.escapeXMLText (value) + quoteChar);
    }

    public String getAttributeValue ()
    {
        return value;
    }
    
    public void setAttributeValue (String value)
    {
        this.value = value;
        if (value.indexOf (quoteChar) != -1)
            quoteChar = 0;

        updateValue ();
    }
    
}