package de.pdark.decentxml;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

public class JoinedInputStream extends InputStream
{
    private List<InputStream> streams = new ArrayList<InputStream> ();
    
    public JoinedInputStream ()
    {
        
    }
    
    public JoinedInputStream add (InputStream in)
    {
        streams.add (in);
        return this;
    }

    @Override
    public int read () throws IOException
    {
        while (!streams.isEmpty ())
        {
            int result = streams.get (0).read ();
            if (result != -1)
                return result;
            
            popStream ();
        }
        
        return -1;
    }

    /**
     * @throws IOException
     */
    private void popStream () throws IOException
    {
        streams.get (0).close ();
        streams.remove (0);
    }

    @Override
    public int read (byte[] b, int off, int len) throws IOException
    {
        while (!streams.isEmpty ())
        {
            int result = streams.get (0).read (b, off, len);
            if (result != -1)
                return result;
            
            popStream ();
        }
        
        return -1;
    }
    
    @Override
    public void close () throws IOException
    {
        while (!streams.isEmpty ())
        {
            popStream ();
        }
        
        // Make this class thow NPEs when being used after closing
        streams = null;
    }
}
