package de.pdark.decentxml;

public class XMLParseException extends RuntimeException
{
    private XMLSource source;
    private int line;
    private int column;
    private int lineStartOffset;
    private int offset;
    
    public XMLParseException (String message, XMLSource source, int offset)
    {
        super (message);
        
        this.source = source;
        if (source != null)
            this.offset = Math.min (offset, source.length ());
        
        //System.out.println (source.substring (offset, source.length ()));
    }
    
    public XMLParseException (String message, XMLTokenizer.Token token)
    {
        this (message, token == null ? null : token.getSource(), token == null ? 0 : token.getStartOffset());
    }
    
    public XMLParseException (String message)
    {
        super (message);
    }
    
    public XMLParseException (String message, Throwable cause)
    {
        super (message, cause);
    }
    
    private void calcLineNumber ()
    {
        line = 1;
        column = 1;
        
        if (source == null)
            return;
        
        lineStartOffset = 0;
        for (int i=0; i<offset; i++)
        {
            char c = source.charAt (i);
            if (c == '\r' || c == '\n')
            {
                line ++;
                if (c == '\r' && i+1 < source.length () && source.charAt (i+1) == '\n')
                    i ++;
                lineStartOffset = i + 1;
            }
        }
        
        column = offset - lineStartOffset + 1;
    }
    
    public XMLSource getSource ()
    {
        return source;
    }
    
    public int getLine ()
    {
        if (line == 0)
            calcLineNumber ();
        
        return line;
    }
    
    public int getColumn ()
    {
        if (line == 0)
            calcLineNumber ();
        
        return column;
    }
    
    public int getOffset ()
    {
        return offset;
    }
    
    public int getLineStartOffset ()
    {
        if (line == 0)
            calcLineNumber ();
        
        return lineStartOffset;
    }
    
    @Override
    public String getMessage ()
    {
        return "Line "+getLine ()+", column " + getColumn () + ": " + super.getMessage ();
    }
}
