package de.pdark.decentxml;


import de.pdark.decentxml.XMLTokenizer.Token;
import de.pdark.decentxml.XMLTokenizer.Type;

public class XMLParser
{
    public static String escapeXMLText (String text)
    {
        return text
            .replaceAll ("&", "&amp;")
            .replaceAll ("<", "&lt;")
            .replaceAll (">", "&gt;")
        ;
    }
    
    public XMLParser ()
    {
        // Do nothing ...
    }
    
    public Document parse (XMLSource source)
    {
        Document doc = new Document ();
        
        XMLTokenizer tokenizer = new XMLTokenizer (source);
        XMLTokenizer.Token token; 

        while ((token = tokenizer.next ()) != null)
        {
            Node n = toNode (token);
            doc.addNode (n);
            
            if (token.getType() == Type.BEGIN_ELEMENT)
            {
                parseElement (tokenizer, (Element)n);
            }
        }
        
        if (doc.getRootElement () == null)
            throw new XMLParseException ("No root element found");
        
        return doc;
    }

    private void parseElement (XMLTokenizer tokenizer, Element parent)
    {
        while (true)
        {
            XMLTokenizer.Token token = tokenizer.next ();
            if (token == null)
                throw new XMLParseException ("Unexpected end-of-file while parsing attributes of element "+parent.getName (), tokenizer.getSource (), tokenizer.getOffset ());
            
            if (token.getType() == Type.BEGIN_ELEMENT_END)
            {
                if ("/>".equals (token.getText ()))
                {
                    parent.setCompactEmpty (true);
                    return;
                }
                
                break;
            }
            else if (token.getType() == Type.ATTRIBUTE || token.getType() == Type.ELEMENT_WHITESPACE)
            {
                parent.addAttribute (toNode (token));
            }
            else  
                // Note: this code will never be executed if the tokenizer works correctly
                throw new XMLParseException ("Unexpected token "+token+" while parsing attributes of element "+parent.getName (), token);
        }
        
        while (true)
        {
            XMLTokenizer.Token token = tokenizer.next ();
            if (token == null)
                throw new XMLParseException ("Unexpected end-of-file while parsing children of element "+parent.getName (), tokenizer.getSource (), tokenizer.getOffset ());
            
            if (token.getType() == Type.END_ELEMENT)
            {
                String endName = token.getText ();
                endName = endName.substring (2, endName.length () - 1);
                String name = endName.trim ();
                
                if (!name.trim ().equals (parent.getName ()))
                    throw new XMLParseException ("End element '"+name+"' doesn't match with '"+parent.getName ()+"'", parent.getStartToken ());
                
                if (endName.length () != parent.getName ().length ())
                    parent.setEndName (endName);
                
                parent.getStartToken ().setEndOffset (token.getEndOffset ());
                
                break;
            }
            
            Node n = toNode (token);
            parent.addChild (n);
            
            if (token.getType() == Type.BEGIN_ELEMENT)
            {
                Element child = (Element)n;
                parseElement (tokenizer, child);
            }
        }
    }

    private Node toNode (Token token)
    {
        switch (token.getType())
        {
        case TEXT: return new Text (token);
        case ATTRIBUTE: return new Attribute (token);
        case BEGIN_ELEMENT: return new Element (token);
        case CDATA: return new Text (token);
        case COMMENT: return new Comment (token);
        case ELEMENT_WHITESPACE: return new Text (token);
        case PROCESSING_INSTRUCTION: return new ProcessingInstruction (token);
        }

        // Note: this code should never be executed. If it is, then there is a new type of Token
        // and the switch wasn't updated for it.
        throw new XMLParseException ("Unexpected token "+token, token);
    }
}

